import torch
from .preprocess import *
from typing import Tuple, List, Union, Optional
from typing_extensions import Literal
from pydantic import BaseModel
import torch as th
import numpy as np

# configs
class Config(BaseModel):
    @property
    def class_name(self):
        return self.__class__.__name__
    
class encoder_cfg(Config):
    # Layer dimensions
    Layer: Tuple[Union[int, str], ...] = (2000,256,64,32)
    # Activation function
    activation: Union[str, None, List[Union[None, str]], Tuple[Union[None, str], ...]] = "relu"

class mvencoder_cfg(Config):
    # atac encoder cfg
    view1_encoder_cfg = encoder_cfg(Layer=(20000,256,64,32))
    # rna encoder cfg
    view2_encoder_cfg = encoder_cfg(Layer=(14,32))
    
class Dataset(Config):
    # Name of the dataset. Must correspond to a filename in data/processed/
    name: str
    # Number of samples to load. Set to None to load all samples
    n_samples: int = None
    # Subset of views to load. Set to None to load all views
    select_views: Tuple[int, ...] = None
    # Subset of labels (classes) to load. Set to None to load all classes
    select_labels: Tuple[int, ...] = None
    # Number of samples to load for each class. Set to None to load all samples
    label_counts: Tuple[int, ...] = None
    # Standard deviation of noise added to the views `noise_views`.
    noise_sd: float = None
    # Subset of views to add noise to
    noise_views: Tuple[int, ...] = None


class Loss_config(Config):
    # Number of views
    n_views: int = 2
    # Number of clusters
    n_clusters: int = None
    # cuda device
    device = 0
    # Terms to use in the loss, separated by '|'. E.g. "ddc_1|ddc_2|ddc_3|" for the DDC clustering loss
    funcs: str
    # Optional weights for the loss terms. Set to None to have all weights equal to 1.
    weights: Tuple[Union[float, int], ...] = None
    # Multiplication factor for the sigma hyperparameter
    rel_sigma = 0.15
    # Tau hyperparameter
    tau = 0.1
    # Delta hyperparameter
    delta = 0.1
    # if zinb_loss
    if_zinb = True
    # Gamma hyperparametre
    gamma = 2.0
    # index for views modeled by ZINB
    zinb_index = list([0])
    # Fraction of batch size to use as the number of negative samples in the contrastive loss. Set to -1 to use all
    # pairs (except the positive) as negative pairs.
    negative_samples_ratio: float = 0.25
    # Similarity function for the contrastive loss. "cos" (default) and "gauss" are supported.
    contrastive_similarity: Literal["cos", "gauss"] = "cos"
    # Enable the adaptive contrastive weighting?
    adaptive_contrastive_weight = True


class Optimizer_config(Config):
    # Base learning rate
    learning_rate: float = 0.0001
    # Max gradient norm for gradient clipping.
    clip_norm: float = 5.0
    # Step size for the learning rate scheduler. None disables the scheduler.
    scheduler_step_size: int = None
    # Multiplication factor for the learning rate scheduler
    scheduler_gamma: float = 0.1


class DDC_config(Config):
    # Number of clusters
    n_clusters: int = None
    # Number of units in the first fully connected layer
    n_hidden = 32
    # Use batch norm after the first fully connected layer?
    use_bn = False
    # If direct or not
    direct = False
    # cuda device
    device = 0

    
class Fusion_config(Config):
    # Fusion method. "mean" constant weights = 1/V. "weighted_mean": Weighted average with learned weights.
    method: Literal["mean", "weighted_mean"]
    # Number of views
    n_views: int

    
class scMVC_simple_config(Config):
    # multiview_encodoers_config
    multiview_encoders_config: mvencoder_cfg = mvencoder_cfg()
    # fusion config
    fusion_config: Fusion_config = Fusion_config(method="weighted_mean", n_views=2)
    # Clustering module config.
    cm_config: Union[DDC_config] = DDC_config(n_clusters=7)
    # Loss function config
    loss_config: Loss_config = Loss_config(n_clusters=7, funcs="ddc_1|ddc_2|ddc_3",)
    # Optimizer config
    optimizer_config = Optimizer_config()

class MLP_config(Config):
    # Shape of the input
    input_size: Tuple[int, ...] = None
    # Units in the network layers
    layers: Tuple[Union[int, str], ...] = (512, 512, 256)
    # Activation function. Can be a single string specifying the activation function for all layers, or a list/tuple of
    # string specifying the activation function for each layer.
    activation: Union[str, None, List[Union[None, str]], Tuple[Union[None, str], ...]] = "relu"
    # Include bias parameters? A single bool for all layers, or a list/tuple of booleans for individual layers.
    use_bias: Union[bool, Tuple[bool, ...]] = True
    # Include batch norm after layers? A single bool for all layers, or a list/tuple of booleans for individual layers.
    use_bn: Union[bool, Tuple[bool, ...]] = False    

class scMVC_contrast_config(Config):
    # multiview_encodoers_config
    multiview_encoders_config: mvencoder_cfg = mvencoder_cfg()
    # Projection head config. Set to None to remove the projection head.
    projector_config: Optional[MLP_config] = None
    # Fusion module config.
    fusion_config:  Fusion_config = Fusion_config(method="weighted_mean", n_views=2)
    # Clustering module config.
    cm_config: Union[DDC_config] = DDC_config(n_clusters=7)
    # Loss function config
    loss_config:  Loss_config = Loss_config(n_clusters=7, funcs="ddc_1|ddc_2|ddc_3|contrast",)
    # Optimizer config
    optimizer_config = Optimizer_config()


# Define a ddc Layer

import torch.nn as nn

class none_layer(nn.Module):
    def __init__(self):
        
        super().__init__()
        
        self.output = None
    
    def forward(self, x):
        self.output = x
        return self.output


class DDC(nn.Module):
    def __init__(self, input_dim, cfg):
        """
        DDC clustering module

        :param input_dim: Shape of inputs.
        :param cfg: DDC config. See `config.defaults.DDC`
        """
        super().__init__()

        hidden_layers = [nn.Linear(input_dim[0], cfg.n_hidden), nn.ReLU()]
        if cfg.use_bn:
            hidden_layers.append(nn.BatchNorm1d(num_features=cfg.n_hidden))
        self.hidden = nn.Sequential(*hidden_layers)
        self.output = nn.Sequential(nn.Linear(cfg.n_hidden, cfg.n_clusters), nn.Softmax(dim=1))
        if cfg.direct:
            self.hidden = none_layer()
            self.output = nn.Sequential(nn.Linear(input_dim[0], cfg.n_clusters), nn.Softmax(dim=1))

    def forward(self, x):
        hidden = self.hidden(x)
        output = self.output(hidden)
        return output, hidden


# kernels for loss calculation 

import torch as th
from torch.nn.functional import relu


EPSILON = 1E-9


def kernel_from_distance_matrix(dist, rel_sigma, min_sigma=EPSILON):

    # `dist` can sometimes contain negative values due to floating point errors, so just set these to zero.
    dist = relu(dist)
    sigma2 = rel_sigma * th.median(dist)
    # Disable gradient for sigma
    sigma2 = sigma2.detach()
    sigma2 = th.where(sigma2 < min_sigma, sigma2.new_tensor(min_sigma), sigma2)
    k = th.exp(- dist / (2 * sigma2))
    return k


def vector_kernel(x, rel_sigma=0.15):
    
    return kernel_from_distance_matrix(cdist(x, x), rel_sigma)


def cdist(X, Y):
    xyT = X @ th.t(Y)
    x2 = th.sum(X**2, dim=1, keepdim=True)
    y2 = th.sum(Y**2, dim=1, keepdim=True)
    d = x2 - 2 * xyT + th.t(y2)
    return d


# comstruct network
# an Encoder for each view of data
import torch.nn as nn
import torch.nn.functional as F
import torch as th

# ZINB loss
class ZINBLoss(nn.Module):
    def __init__(self):
        super(ZINBLoss, self).__init__()

    def forward(self, x, mean, disp, pi, scale_factor=1.0, ridge_lambda=0.0):
        eps = 1e-10
        scale_factor = scale_factor[:, None]
        mean = mean * scale_factor
        
        t1 = th.lgamma(disp+eps) + th.lgamma(x+1.0) - th.lgamma(x+disp+eps)
        t2 = (disp+x) * th.log(1.0 + (mean/(disp+eps))) + (x * (th.log(disp+eps) - th.log(mean+eps)))
        nb_final = t1 + t2

        nb_case = nb_final - th.log(1.0-pi+eps)
        zero_nb = th.pow(disp/(disp+mean+eps), disp)
        zero_case = -th.log(pi + ((1.0-pi)*zero_nb)+eps)
        result = th.where(th.le(x, 1e-8), zero_case, nb_case)
        
        if ridge_lambda > 0:
            ridge = ridge_lambda*th.square(pi)
            result += ridge
        
        result = th.mean(result)
        return result

    
# basic layer functions
class GaussianNoise(nn.Module):
    def __init__(self, sigma=0):
        super(GaussianNoise, self).__init__()
        self.sigma = sigma
    
    def forward(self, x):
        if self.training:
            x = x + self.sigma * th.randn_like(x)
        return x

class MeanAct(nn.Module):
    def __init__(self):
        super(MeanAct, self).__init__()

    def forward(self, x):
        return th.clamp(th.exp(x), min=1e-5, max=1e6)

class DispAct(nn.Module):
    def __init__(self):
        super(DispAct, self).__init__()

    def forward(self, x):
        return th.clamp(F.softplus(x), min=1e-4, max=1e4)

def buildNetwork(layers, type, activation="relu"):
    net = []
    for i in range(1, len(layers)):
        net.append(nn.Linear(layers[i-1], layers[i]))
        if activation=="relu":
            net.append(nn.ReLU())
        elif activation=="sigmoid":
            net.append(nn.Sigmoid())
    return nn.Sequential(*net)

class Backbone(nn.Module):
    def __init__(self):
        """
        Backbone base class
        """
        super().__init__()
        self.layers = nn.ModuleList()

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        return x
    
# define multi layer perceptron class for projection
class MLP(Backbone):
    def __init__(self, cfg, input_size=None, **_):
        """
        MLP backbone

        :param cfg: MLP config
        :type cfg: config.defaults.MLP
        :param input_size: Optional input size which overrides the one set in `cfg`.
        :type input_size: Optional[Union[List, Tuple]]
        :param _:
        :type _:
        """
        super().__init__()
        self.output_size = self.create_linear_layers(cfg, self.layers, input_size=input_size)

    @staticmethod
    def get_activation_module(a):
        if a == "relu":
            return nn.ReLU()
        elif a == "sigmoid":
            return nn.Sigmoid()
        elif a == "tanh":
            return nn.Tanh()
        elif a == "softmax":
            return nn.Softmax(dim=1)
        elif a.startswith("leaky_relu"):
            neg_slope = float(a.split(":")[1])
            return nn.LeakyReLU(neg_slope)
        else:
            raise RuntimeError(f"Invalid MLP activation: {a}.")

    @classmethod
    def create_linear_layers(cls, cfg, layer_container, input_size=None):
        # `input_size` takes priority over `cfg.input_size`
        if input_size is not None:
            output_size = list(input_size)
        else:
            output_size = list(cfg.input_size)

        if len(output_size) > 1:
            layer_container.append(nn.Flatten())
            output_size = [np.prod(output_size)]

        n_layers = len(cfg.layers)
        activations = helpers.ensure_iterable(cfg.activation, expected_length=n_layers)
        use_bias = helpers.ensure_iterable(cfg.use_bias, expected_length=n_layers)
        use_bn = helpers.ensure_iterable(cfg.use_bn, expected_length=n_layers)

        for n_units, act, _use_bias, _use_bn in zip(cfg.layers, activations, use_bias, use_bn):
            # If we get n_units = -1, then the number of units should be the same as the previous number of units, or
            # the input dim.
            if n_units == -1:
                n_units = output_size[0]

            layer_container.append(nn.Linear(in_features=output_size[0], out_features=n_units, bias=_use_bias))
            if _use_bn:
                # Add BN before activation
                layer_container.append(nn.BatchNorm1d(num_features=n_units))
            if act is not None:
                # Add activation
                layer_container.append(cls.get_activation_module(act))
            output_size[0] = n_units

        return output_size
    

# encoder function
class Autoencoder(nn.Module):
    
    def __init__(self,cfg):
        """cfg: Layer = [Input_dim, ..., Latent_dim]
                activation = "ReLU"
        """ 
        super(Autoencoder, self).__init__()
        self.Layer = cfg.Layer
        self.activation = cfg.activation
        self.encoder = buildNetwork(self.Layer[:-1], type="encode", activation=self.activation)
        self.decoder = buildNetwork(self.Layer[:0:-1], type="decode", activation=self.activation)
        self._enc_mu = nn.Linear(self.Layer[-2], self.Layer[-1])
        self._dec_mean = nn.Sequential(nn.Linear(self.Layer[1], self.Layer[0]), MeanAct())
        self._dec_disp = nn.Sequential(nn.Linear(self.Layer[1], self.Layer[0]), DispAct())
        self._dec_pi = nn.Sequential(nn.Linear(self.Layer[1], self.Layer[0]), nn.Sigmoid())
        self.zinb_loss = ZINBLoss().cuda()
    
    def output_size(self):
        return self.Layer[-1]
      
    def forward(self, x):
        h = self.encoder(x+th.randn_like(x) * 1.5)
        z = self._enc_mu(h)
        h = self.decoder(z)
        _mean = self._dec_mean(h)
        _disp = self._dec_disp(h)
        _pi = self._dec_pi(h)

        h0 = self.encoder(x)
        z0 = self._enc_mu(h0)
        return [z0, _mean, _disp, _pi]
 

# multiview autoencoder function
class multiview_autoencoders(nn.Module):
    
    def __init__(self, mvae_configs):

        super().__init__()

        self.mvae = nn.ModuleList()
        for cfg in mvae_configs: 
            self.mvae.append(Autoencoder(cfg[1]))

    @property
    def output_sizes(self):
        return [ec.output_size() for ec in self.mvae]


    def forward(self, mv_input):
        views = mv_input[0]
        assert len(views) == len(self.mvae)
        outputs_latent = [ae(v)[0] for ae, v in zip(self.mvae, views)]
        outputs_zinb_mean = [ae(v)[1] for ae, v in zip(self.mvae, views)]
        outputs_zinb_disp = [ae(v)[2] for ae, v in zip(self.mvae, views)]
        outputs_zinb_pi = [ae(v)[3] for ae, v in zip(self.mvae, views)]
        return [outputs_latent, outputs_zinb_mean, outputs_zinb_disp, outputs_zinb_pi]


from .utils import *
import h5py
import scipy as sp
import numpy as np
import scanpy as sc
import pandas as pd
from scipy.sparse import*
import random
# define class sc-sequencing dataset
import torch
import torch.utils.data as data


def normalize(adata, copy=True, highly_genes=False, filter_min_counts=False, size_factors=True, normalize_input=True, logtrans_input=True):
    if isinstance(adata, sc.AnnData):
        if copy:
            adata = adata.copy()
    elif isinstance(adata, str):
        adata = sc.read(adata)
    else:
        raise NotImplementedError
    """
    norm_error = 'Make sure that the dataset (adata.X) contains unnormalized count data.'
    assert 'n_count' not in adata.obs, norm_error
    if adata.X.size < 50e6: # check if adata.X is integer only if array is small
        if sp.sparse.issparse(adata.X):
            assert (adata.X.astype(int) != adata.X).nnz == 0, norm_error
        else:
            assert np.all(adata.X.astype(int) == adata.X), norm_error
    """
    if filter_min_counts:
        sc.pp.filter_genes(adata, min_counts=0)
        
    if logtrans_input:
        sc.pp.log1p(adata)
        
    if highly_genes != None:
        sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5, n_top_genes = highly_genes, subset=True)
        
    if size_factors or normalize_input or logtrans_input:
        adata.raw = adata.copy()
    else:
        adata.raw = adata
    if size_factors:
        sc.pp.normalize_per_cell(adata,min_counts=0)
        adata.obs['size_factors'] = adata.obs.n_counts / np.median(adata.obs.n_counts)
    else:
        adata.obs['size_factors'] = 1.0
    if normalize_input:
        sc.pp.scale(adata)
    return adata

class multiviewDataset(data.Dataset):
    
    # highly_genes: [[a1,a2...],[b1,b2...]], ai the views needs to select, bi the number of highly var genes
    def __init__(self, views, labels, device, if_normalize=True, highly_genes=[[0],[4000]]):
        
        self.n_views = len(views)
        self.views = views
        self.labels = labels
        self.device = device
        # we only normalize select highly var genes for view[highly_genes[0]]
        highly_cnt = 0 # count the number of views needs hvg
        if if_normalize:
            for i in range(self.n_views):
                self.views[i] = sc.AnnData(self.views[i])
                if i in list(highly_genes[0]):
                    self.views[i] = normalize(self.views[i], highly_genes=highly_genes[1][highly_cnt])
                    highly_cnt += 1
                else:
                    self.views[i] = normalize(self.views[i], copy=True, highly_genes=False, filter_min_counts=False,
                                              size_factors=True, normalize_input=True, logtrans_input=True)

                print("The dimension of view %d is %d"%(i, np.shape(self.views[i].X)[1]))
            
    def __getitem__(self, index):

        batch_cells = [torch.Tensor(v.X[index]).to(self.device) for v in self.views]
        batch_raw = [torch.Tensor(v.raw.X[index]).to(self.device) for v in self.views]
        batch_sf = [torch.tensor(v.obs['size_factors'][index]).to(self.device) for v in self.views]
        batch_labels = torch.tensor(self.labels[index]).to(self.device)
       
        return [batch_cells, batch_raw, batch_sf], batch_labels
    
    def __len__(self):
        return len(self.labels)
    
class multiviewDataset_nolabel(data.Dataset):
    
    # highly_genes: [[a1,a2...],[b1,b2...]], ai the views needs to select, bi the number of highly var genes
    def __init__(self, views, device, if_normalize=True, highly_genes=[[0],[4000]]):
        
        self.n_views = len(views)
        self.views = views
        self.device = device
        # we only normalize select highly var genes for view[highly_genes[0]]
        highly_cnt = 0 # count the number of views needs hvg
        if if_normalize:
            for i in range(self.n_views):
                self.views[i] = sc.AnnData(self.views[i])
                if i in list(highly_genes[0]):
                    self.views[i] = normalize(self.views[i], highly_genes=highly_genes[1][highly_cnt])
                    highly_cnt += 1
                else:
                    self.views[i] = normalize(self.views[i], copy=True, highly_genes=False, filter_min_counts=False,
                                              size_factors=True, normalize_input=True, logtrans_input=True)

                print("The dimension of view %d is %d"%(i, np.shape(self.views[i].X)[1]))
            
    def __getitem__(self, index):

        batch_cells = [torch.Tensor(v.X[index]).to(self.device) for v in self.views]
        batch_raw = [torch.Tensor(v.raw.X[index]).to(self.device) for v in self.views]
        batch_sf = [torch.tensor(v.obs['size_factors'][index]).to(self.device) for v in self.views]
       
        return [batch_cells, batch_raw, batch_sf]
    
    def __len__(self):
        return len(self.views[1])


def read_clean(data):
    assert isinstance(data, np.ndarray)
    if data.dtype.type is np.bytes_:
        data = decode(data)
    if data.size == 1:
        data = data.flat[0]
    return data


def dict_from_group(group):
    assert isinstance(group, h5py.Group)
    d = dotdict()
    for key in group:
        if isinstance(group[key], h5py.Group):
            value = dict_from_group(group[key])
        else:
            value = read_clean(group[key][...])
        d[key] = value
    return d


def read_data(filename, sparsify = False, skip_exprs = False):
    with h5py.File(filename, "r") as f:
        obs = pd.DataFrame(dict_from_group(f["obs"]), index = decode(f["obs_names"][...]))
        var = pd.DataFrame(dict_from_group(f["var"]), index = decode(f["var_names"][...]))
        uns = dict_from_group(f["uns"])
        if not skip_exprs:
            exprs_handle = f["exprs"]
            if isinstance(exprs_handle, h5py.Group):
                mat = sp.sparse.csr_matrix((exprs_handle["data"][...], exprs_handle["indices"][...],
                                               exprs_handle["indptr"][...]), shape = exprs_handle["shape"][...])
            else:
                mat = exprs_handle[...].astype(np.float32)
                if sparsify:
                    mat = sp.sparse.csr_matrix(mat)
        else:
            mat = sp.sparse.csr_matrix((obs.shape[0], var.shape[0]))
    return mat, var, uns


# define a Fusion Layer

import numpy as np
import torch as th
import torch.nn as nn


class _Fusion(nn.Module):
    def __init__(self, cfg, input_sizes):

        super().__init__()
        self.cfg = cfg
        self.input_sizes = input_sizes
        self.output_size = None

    def forward(self, inputs):
        raise NotImplementedError()

    @classmethod
    def get_weighted_sum_output_size(cls, input_sizes):
        flat_sizes = [np.prod(s) for s in input_sizes]
        assert all(s == flat_sizes[0] for s in flat_sizes), f"Fusion method {cls.__name__} requires the flat output" \
                                                            f" shape from all backbones to be identical." \
                                                            f" Got sizes: {input_sizes} -> {flat_sizes}."
        return [flat_sizes[0]]

    def get_weights(self, softmax=True):
        out = []
        if hasattr(self, "weights"):
            out = self.weights
            if softmax:
                out = nn.functional.softmax(self.weights, dim=-1)
        return out

    def update_weights(self, inputs, a):
        pass


class Mean(_Fusion):
    def __init__(self, cfg, input_sizes):
        """
        Mean fusion.

        :param cfg: Fusion config. See config.defaults.Fusion
        :param input_sizes: Input shapes
        """
        super().__init__(cfg, input_sizes)
        self.output_size = self.get_weighted_sum_output_size(input_sizes)

    def forward(self, inputs):
        return th.mean(th.stack(inputs, -1), dim=-1)


class WeightedMean(_Fusion):
    """
    Weighted mean fusion.

    :param cfg: Fusion config. See config.defaults.Fusion
    :param input_sizes: Input shapes
    """
    def __init__(self, cfg, input_sizes):
        super().__init__(cfg, input_sizes)
        self.weights = nn.Parameter(th.full((self.cfg.n_views,), 1 / self.cfg.n_views), requires_grad=True)
        self.output_size = self.get_weighted_sum_output_size(input_sizes)

    def forward(self, inputs):
        return _weighted_sum(inputs, self.weights, normalize_weights=True)


def _weighted_sum(tensors, weights, normalize_weights=True):
    if normalize_weights:
        weights = nn.functional.softmax(weights, dim=0)
    out = th.sum(weights[None, None, :] * th.stack(tensors, dim=-1), dim=-1)
    return out


MODULES = {
    "mean": Mean,
    "weighted_mean": WeightedMean,
}


def get_fusion_module(cfg, input_sizes):
    return MODULES[cfg.method](cfg, input_sizes)


import torch as th
import torch.nn as nn
import torch.nn.functional as F
from .loss_kernels import *

DATETIME_FMT = "%Y-%m-%d_%H-%M-%S"

EPSILON = 1E-9
DEBUG_MODE = False


def triu(X):
    # Sum of strictly upper triangular part
    return th.sum(th.triu(X, diagonal=1))


def _atleast_epsilon(X, eps=EPSILON):
    """
    Ensure that all elements are >= `eps`.

    :param X: Input elements
    :type X: th.Tensor
    :param eps: epsilon
    :type eps: float
    :return: New version of X where elements smaller than `eps` have been replaced with `eps`.
    :rtype: th.Tensor
    """
    return th.where(X < eps, X.new_tensor(eps), X)


def d_cs(A, K, n_clusters):
    """
    Cauchy-Schwarz divergence.

    :param A: Cluster assignment matrix
    :type A:  th.Tensor
    :param K: Kernel matrix
    :type K: th.Tensor
    :param n_clusters: Number of clusters
    :type n_clusters: int
    :return: CS-divergence
    :rtype: th.Tensor
    """
    nom = th.t(A) @ K @ A
    dnom_squared = th.unsqueeze(th.diagonal(nom), -1) @ th.unsqueeze(th.diagonal(nom), 0)

    nom = _atleast_epsilon(nom)
    dnom_squared = _atleast_epsilon(dnom_squared, eps=EPSILON**2)

    d = 2 / (n_clusters * (n_clusters - 1)) * triu(nom / th.sqrt(dnom_squared))
    return d


# ======================================================================================================================
# Loss terms
# ======================================================================================================================

class LossTerm:
    # Names of tensors required for the loss computation
    required_tensors = []

    def __init__(self, *args, **kwargs):
        """
        Base class for a term in the loss function.

        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        """
        pass

    def __call__(self, net, cfg, extra):
        raise NotImplementedError()


class DDC1(LossTerm):
    """
    L_1 loss from DDC
    """
    required_tensors = ["hidden_kernel"]

    def __call__(self, net, cfg, extra):
        return d_cs(net.output[0], extra["hidden_kernel"], cfg.n_clusters)


class DDC2(LossTerm):
    """
    L_2 loss from DDC
    """
    def __call__(self, net, cfg, extra):
        n = net.output[0].size(0)
        return 2 / (n * (n - 1)) * triu(net.output[0] @ th.t(net.output[0]))


class DDC2Flipped(LossTerm):
    """
    Flipped version of the L_2 loss from DDC. Used by EAMC
    """

    def __call__(self, net, cfg, extra):
        return 2 / (cfg.n_clusters * (cfg.n_clusters - 1)) * triu(th.t(net.output[0]) @ net.output[0])


class DDC3(LossTerm):
    """
    L_3 loss from DDC
    """
    required_tensors = ["hidden_kernel"]

    def __init__(self, cfg):
        super().__init__()
        self.eye = th.eye(cfg.n_clusters, device=th.device('cuda',cfg.device))

    def __call__(self, net, cfg, extra):
        m = th.exp(-cdist(net.output[0], self.eye))
        return d_cs(m, extra["hidden_kernel"], cfg.n_clusters)

class kl_div(LossTerm):
    """
    KL divergence loss to avoid cluster collapse
    """
    def __call__(self, net, cfg, extra):
        q = th.ones(net.output[0].shape[1], device=th.device('cuda',cfg.device))/cfg.n_clusters
        return F.kl_div(th.mean(net.output[0], dim=0), q, reduction='batchmean')
    
class zinb1(LossTerm):
    """
    zinb loss for highly-sparse rna data
    """
    def __call__(self, net, cfg, extra):

        x = net.mv_input[1][0]
        mean = net.output[2][0]
        disp = net.output[3][0]
        pi = net.output[4][0]
        scale_factor = net.mv_input[2][0][:, None]
        ridge_lambda=0.0
        eps = 1e-10
        mean = mean * scale_factor
        
        t1 = th.lgamma(disp+eps) + th.lgamma(x+1.0) - th.lgamma(x+disp+eps)
        t2 = (disp+x) * th.log(1.0 + (mean/(disp+eps))) + (x * (th.log(disp+eps) - th.log(mean+eps)))
        nb_final = t1 + t2

        nb_case = nb_final - th.log(1.0-pi+eps)
        zero_nb = th.pow(disp/(disp+mean+eps), disp)
        zero_case = -th.log(pi + ((1.0-pi)*zero_nb)+eps)
        result = th.where(th.le(x, 1e-8), zero_case, nb_case)
        
        if ridge_lambda > 0:
            ridge = ridge_lambda*th.square(pi)
            result += ridge
        
        result = th.mean(result)
        return result

class zinb2(LossTerm):
    """
    zinb loss for highly-sparse rna data
    """
    def __call__(self, net, cfg, extra):

        x = net.mv_input[1][1]
        mean = net.output[2][1]
        disp = net.output[3][1]
        pi = net.output[4][1]
        scale_factor = net.mv_input[2][1][:, None]
        ridge_lambda=0.0
        eps = 1e-10
        mean = mean * scale_factor
        
        t1 = th.lgamma(disp+eps) + th.lgamma(x+1.0) - th.lgamma(x+disp+eps)
        t2 = (disp+x) * th.log(1.0 + (mean/(disp+eps))) + (x * (th.log(disp+eps) - th.log(mean+eps)))
        nb_final = t1 + t2

        nb_case = nb_final - th.log(1.0-pi+eps)
        zero_nb = th.pow(disp/(disp+mean+eps), disp)
        zero_case = -th.log(pi + ((1.0-pi)*zero_nb)+eps)
        result = th.where(th.le(x, 1e-8), zero_case, nb_case)
        
        if ridge_lambda > 0:
            ridge = ridge_lambda*th.square(pi)
            result += ridge
        
        result = th.mean(result)
        return result

class Contrastive(LossTerm):
    large_num = 1e9

    def __init__(self, cfg):
        """
        Contrastive loss function

        :param cfg: Loss function config
        :type cfg: config.defaults.Loss
        """
        super().__init__()
        # Select which implementation to use
        if cfg.negative_samples_ratio == -1:
            self._loss_func = self._loss_without_negative_sampling
        else:
            self.eye = th.eye(cfg.n_clusters, device=th.device('cuda',cfg.device))
            self._loss_func = self._loss_with_negative_sampling

        # Set similarity function
        if cfg.contrastive_similarity == "cos":
            self.similarity_func = self._cosine_similarity
        elif cfg.contrastive_similarity == "gauss":
            self.similarity_func = vector_kernel
        else:
            raise RuntimeError(f"Invalid contrastive similarity: {cfg.contrastive_similarity}")

    @staticmethod
    def _norm(mat):
        return th.nn.functional.normalize(mat, p=2, dim=1)

    @staticmethod
    def get_weight(net):
        w = th.min(th.nn.functional.softmax(net.fusion.weights.detach(), dim=0))
        return w

    @classmethod
    def _normalized_projections(cls, net):
        n = net.projections.size(0) // 2
        h1, h2 = net.projections[:n], net.projections[n:]
        h2 = cls._norm(h2)
        h1 = cls._norm(h1)
        return n, h1, h2

    @classmethod
    def _cosine_similarity(cls, projections):
        h = cls._norm(projections)
        return h @ h.t()

    def _draw_negative_samples(self, net, cfg, v, pos_indices):
        """
        Construct set of negative samples.

        :param net: Model
        :type net: Union[models.simple_mvc.SiMVC, models.contrastive_mvc.CoMVC]
        :param cfg: Loss config
        :type cfg: config.defaults.Loss
        :param v: Number of views
        :type v: int
        :param pos_indices: Row indices of the positive samples in the concatenated similarity matrix
        :type pos_indices: th.Tensor
        :return: Indices of negative samples
        :rtype: th.Tensor
        """
        cat = net.output[0].detach().argmax(dim=1)
        cat = th.cat(v * [cat], dim=0)

        weights = (1 - self.eye[cat])[:, cat[[pos_indices]]].T
        n_negative_samples = int(cfg.negative_samples_ratio * cat.size(0))
        negative_sample_indices = th.multinomial(weights, n_negative_samples, replacement=True)
        if DEBUG_MODE:
            self._check_negative_samples_valid(cat, pos_indices, negative_sample_indices)
        return negative_sample_indices

    @staticmethod
    def _check_negative_samples_valid(cat, pos_indices, neg_indices):
        pos_cats = cat[pos_indices].view(-1, 1)
        neg_cats = cat[neg_indices]
        assert (pos_cats != neg_cats).detach().cpu().numpy().all()

    @staticmethod
    def _get_positive_samples(logits, v, n):
        """
        Get positive samples

        :param logits: Input similarities
        :type logits: th.Tensor
        :param v: Number of views
        :type v: int
        :param n: Number of samples per view (batch size)
        :type n: int
        :return: Similarities of positive pairs, and their indices
        :rtype: Tuple[th.Tensor, th.Tensor]
        """
        diagonals = []
        inds = []
        for i in range(1, v):
            diagonal_offset = i * n
            diag_length = (v - i) * n
            _upper = th.diagonal(logits, offset=diagonal_offset)
            _lower = th.diagonal(logits, offset=-1 * diagonal_offset)
            _upper_inds = th.arange(0, diag_length)
            _lower_inds = th.arange(i * n, v * n)
            if DEBUG_MODE:
                assert _upper.size() == _lower.size() == _upper_inds.size() == _lower_inds.size() == (diag_length,)
            diagonals += [_upper, _lower]
            inds += [_upper_inds, _lower_inds]

        pos = th.cat(diagonals, dim=0)
        pos_inds = th.cat(inds, dim=0)
        return pos, pos_inds

    def _loss_with_negative_sampling(self, net, cfg, extra):
        """
        Contrastive loss implementation with negative sampling.

        :param net: Model
        :type net: Union[models.simple_mvc.SiMVC, models.contrastive_mvc.CoMVC]
        :param cfg: Loss config
        :type cfg: config.defaults.Loss
        :param extra:
        :type extra:
        :return: Loss value
        :rtype: th.Tensor
        """
        n = net.output[0].size(0)
        v = len(net.output[1])
        logits = self.similarity_func(net.projections) / cfg.tau

        pos, pos_inds = self._get_positive_samples(logits, v, n)
        neg_inds = self._draw_negative_samples(net, cfg, v, pos_inds)
        neg = logits[pos_inds.view(-1, 1), neg_inds]

        inputs = th.cat((pos.view(-1, 1), neg), dim=1)
        labels = th.zeros(v * (v - 1) * n, device=th.device('cuda',cfg.device), dtype=th.long)
        loss = th.nn.functional.cross_entropy(inputs, labels)

        if cfg.adaptive_contrastive_weight:
            loss *= self.get_weight(net)

        return cfg.delta * loss

    def _loss_without_negative_sampling(self, net, cfg, extra):
        """
        Contrastive loss implementation without negative sampling.
        Adapted from: https://github.com/google-research/simclr/blob/master/objective.py

        :param net: Model
        :type net: Union[models.simple_mvc.SiMVC, models.contrastive_mvc.CoMVC]
        :param cfg: Loss config
        :type cfg: config.defaults.Loss
        :param extra:
        :type extra:
        :return:
        :rtype:
        """
        assert len(net.output[1]) == 2, "Contrastive loss without negative sampling only supports 2 views."
        n, h1, h2 = self._normalized_projections(net)

        labels = th.arange(0, n, device=th.device('cuda',cfg.device), dtype=th.long)
        masks = th.eye(n, device=th.device('cuda',cfg.device))

        logits_aa = ((h1 @ h1.t()) / cfg.tau) - masks * self.large_num
        logits_bb = ((h2 @ h2.t()) / cfg.tau) - masks * self.large_num

        logits_ab = (h1 @ h2.t()) / cfg.tau
        logits_ba = (h2 @ h1.t()) / cfg.tau

        loss_a = th.nn.functional.cross_entropy(th.cat((logits_ab, logits_aa), dim=1), labels)
        loss_b = th.nn.functional.cross_entropy(th.cat((logits_ba, logits_bb), dim=1), labels)

        loss = (loss_a + loss_b)

        if cfg.adaptive_contrastive_weight:
            loss *= self.get_weight(net)

        return cfg.delta * loss

    def __call__(self, net, cfg, extra):
        return self._loss_func(net, cfg, extra)


# ======================================================================================================================
# Extra functions
# ======================================================================================================================

def hidden_kernel(net, cfg):
    return vector_kernel(net.hidden, cfg.rel_sigma)


# ======================================================================================================================
# Loss class
# ======================================================================================================================

class Loss(nn.Module):
    # Possible terms to include in the loss
    TERM_CLASSES = {
        "ddc_1": DDC1,
        "ddc_2": DDC2,
        "ddc_2_flipped": DDC2Flipped,
        "ddc_3": DDC3,
        "kl_div":kl_div,
        "zinb_1":zinb1,
        "zinb_2":zinb2,
        "contrast": Contrastive,
    }
    # Functions to compute the required tensors for the terms.
    EXTRA_FUNCS = {
        "hidden_kernel": hidden_kernel,
    }

    def __init__(self, cfg):
        """
        Implementation of a general loss function

        :param cfg: Loss function config
        :type cfg: config.defaults.Loss
        """
        super().__init__()
        self.cfg = cfg

        self.names = cfg.funcs.split("|")
        self.weights = cfg.weights if cfg.weights is not None else len(self.names) * [1]

        self.terms = []
        for term_name in self.names:
            self.terms.append(self.TERM_CLASSES[term_name](cfg))

        self.required_extras_names = list(set(sum([t.required_tensors for t in self.terms], [])))

    def forward(self, net, ignore_in_total=tuple()):
        extra = {name: self.EXTRA_FUNCS[name](net, self.cfg) for name in self.required_extras_names}
        loss_values = {}
        for name, term, weight in zip(self.names, self.terms, self.weights):
            value = term(net, self.cfg, extra)
            # If we got a dict, add each term from the dict with "name/" as the scope.
            if isinstance(value, dict):
                for key, _value in value.items():
                    loss_values[f"{name}/{key}"] = weight * _value
            # Otherwise, just add the value to the dict directly
            else:
                loss_values[name] = weight * value

        loss_values["tot"] = sum([loss_values[k] for k in loss_values.keys() if k not in ignore_in_total])

        return loss_values


import torch as th


class Optimizer:
    def __init__(self, cfg, params):
        """
        Wrapper class for optimizers

        :param cfg: Optimizer config
        :type cfg: config.defaults.Optimizer
        :param params: Parameters to associate with the optimizer
        :type params:
        """
        self.clip_norm = cfg.clip_norm
        self.params = params
        self._opt = th.optim.Adam(params, lr=cfg.learning_rate)
        if cfg.scheduler_step_size is not None:
            assert cfg.scheduler_gamma is not None
            self._sch = th.optim.lr_scheduler.StepLR(self._opt, step_size=cfg.scheduler_step_size,
                                                     gamma=cfg.scheduler_gamma)
        else:
            self._sch = None

    def zero_grad(self):
        return self._opt.zero_grad()

    def step(self, epoch):
        if self._sch is not None:
            # Only step the scheduler at integer epochs, and don't step on the first epoch.
            if epoch.is_integer() and epoch > 0:
                self._sch.step()

        if self.clip_norm is not None:
            th.nn.utils.clip_grad_norm_(self.params, self.clip_norm)

        out = self._opt.step()
        return out



import json
import functools
import operator
import collections
import jgraph
import numpy as np
import scipy.sparse
import tqdm


class dotdict(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def in_ipynb():  # pragma: no cover
    try:
        # noinspection PyUnresolvedReferences
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True   # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter


def smart_tqdm():  # pragma: no cover
    if in_ipynb():
        return tqdm.tqdm_notebook
    return tqdm.tqdm


def with_self_graph(fn):
    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        with self.graph.as_default():
            return fn(self, *args, **kwargs)
    return wrapped


# Wraps a batch function into minibatch version
def minibatch(batch_size, desc, use_last=False, progress_bar=True):
    def minibatch_wrapper(func):
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            total_size = args[0].shape[0]
            if use_last:
                n_batch = np.ceil(
                    total_size / float(batch_size)
                ).astype(np.int)
            else:
                n_batch = max(1, np.floor(
                    total_size / float(batch_size)
                ).astype(np.int))
            for batch_idx in smart_tqdm()(
                range(n_batch), desc=desc, unit="batches",
                leave=False, disable=not progress_bar
            ):
                start = batch_idx * batch_size
                end = min((batch_idx + 1) * batch_size, total_size)
                this_args = (item[start:end] for item in args)
                func(*this_args, **kwargs)
        return wrapped_func
    return minibatch_wrapper


# Avoid sklearn warning
def encode_integer(label, sort=False):
    label = np.array(label).ravel()
    classes = np.unique(label)
    if sort:
        classes.sort()
    mapping = {v: i for i, v in enumerate(classes)}
    return np.array([mapping[v] for v in label]), classes


# Avoid sklearn warning
def encode_onehot(label, sort=False, ignore=None):
    i, c = encode_integer(label, sort)
    onehot = scipy.sparse.csc_matrix((
        np.ones_like(i, dtype=np.int32), (np.arange(i.size), i)
    ))
    if ignore is None:
        ignore = []
    return onehot[:, ~np.in1d(c, ignore)].tocsr()


class CellTypeDAG(object):

    def __init__(self, graph=None, vdict=None):
        self.graph = jgraph.Graph(directed=True) if graph is None else graph
        self.vdict = {} if vdict is None else vdict

    @classmethod
    def load(cls, file):
        if file.endswith(".json"):
            return cls.load_json(file)
        elif file.endswith(".obo"):
            return cls.load_obo(file)
        else:
            raise ValueError("Unexpected file format!")

    @classmethod
    def load_json(cls, file):
        with open(file, "r") as f:
            d = json.load(f)
        dag = cls()
        dag._build_tree(d)
        return dag

    @classmethod
    def load_obo(cls, file):  # Only building on "is_a" relation between CL terms
        import pronto
        ont = pronto.Ontology(file)
        graph, vdict = jgraph.Graph(directed=True), {}
        for item in ont:
            if not item.id.startswith("CL"):
                continue
            if "is_obsolete" in item.other and item.other["is_obsolete"][0] == "true":
                continue
            graph.add_vertex(
                name=item.id, cell_ontology_class=item.name,
                desc=str(item.desc), synonyms=[(
                    "%s (%s)" % (syn.desc, syn.scope)
                 ) for syn in item.synonyms]
            )
            assert item.id not in vdict
            vdict[item.id] = item.id
            assert item.name not in vdict
            vdict[item.name] = item.id
            for synonym in item.synonyms:
                if synonym.scope == "EXACT" and synonym.desc != item.name:
                    vdict[synonym.desc] = item.id
        for source in graph.vs:
            for relation in ont[source["name"]].relations:
                if relation.obo_name != "is_a":
                    continue
                for target in ont[source["name"]].relations[relation]:
                    if not target.id.startswith("CL"):
                        continue
                    graph.add_edge(
                        source["name"],
                        graph.vs.find(name=target.id.split()[0])["name"]
                    )
                    # Split because there are many "{is_infered...}" suffix,
                    # falsely joined to the actual id when pronto parses the
                    # obo file
        return cls(graph, vdict)

    def _build_tree(self, d, parent=None):  # For json loading
        self.graph.add_vertex(name=d["name"])
        v = self.graph.vs.find(d["name"])
        if parent is not None:
            self.graph.add_edge(v, parent)
        self.vdict[d["name"]] = d["name"]
        if "alias" in d:
            for alias in d["alias"]:
                self.vdict[alias] = d["name"]
        if "children" in d:
            for subd in d["children"]:
                self._build_tree(subd, v)

    def get_vertex(self, name):
        return self.graph.vs.find(self.vdict[name])

    def is_related(self, name1, name2):
        return self.is_descendant_of(name1, name2) \
            or self.is_ancestor_of(name1, name2)

    def is_descendant_of(self, name1, name2):
        if name1 not in self.vdict or name2 not in self.vdict:
            return False
        shortest_path = self.graph.shortest_paths(
            self.get_vertex(name1), self.get_vertex(name2)
        )[0][0]
        return np.isfinite(shortest_path)

    def is_ancestor_of(self, name1, name2):
        if name1 not in self.vdict or name2 not in self.vdict:
            return False
        shortest_path = self.graph.shortest_paths(
            self.get_vertex(name2), self.get_vertex(name1)
        )[0][0]
        return np.isfinite(shortest_path)

    def conditional_prob(self, name1, name2):  # p(name1|name2)
        if name1 not in self.vdict or name2 not in self.vdict:
            return 0
        self.graph.vs["prob"] = 0
        v2_parents = list(self.graph.bfsiter(
            self.get_vertex(name2), mode=jgraph.OUT))
        v1_parents = list(self.graph.bfsiter(
            self.get_vertex(name1), mode=jgraph.OUT))
        for v in v2_parents:
            v["prob"] = 1
        while True:
            changed = False
            for v1_parent in v1_parents[::-1]:  # Reverse may be more efficient
                if v1_parent["prob"] != 0:
                    continue
                v1_parent["prob"] = np.prod([
                    v["prob"] / v.degree(mode=jgraph.IN)
                    for v in v1_parent.neighbors(mode=jgraph.OUT)
                ])
                if v1_parent["prob"] != 0:
                    changed = True
            if not changed:
                break
        return self.get_vertex(name1)["prob"]

    def similarity(self, name1, name2, method="probability"):
        if method == "probability":
            return (
                self.conditional_prob(name1, name2) +
                self.conditional_prob(name2, name1)
            ) / 2
        # if method == "distance":
        #     return self.distance_ratio(name1, name2)
        raise ValueError("Invalid method!")  # pragma: no cover

    def count_reset(self):
        self.graph.vs["raw_count"] = 0
        self.graph.vs["prop_count"] = 0  # count propagated from children
        self.graph.vs["count"] = 0

    def count_set(self, name, count):
        self.get_vertex(name)["raw_count"] = count

    def count_update(self):
        origins = [v for v in self.graph.vs.select(raw_count_gt=0)]
        for origin in origins:
            for v in self.graph.bfsiter(origin, mode=jgraph.OUT):
                if v != origin:  # bfsiter includes the vertex self
                    v["prop_count"] += origin["raw_count"]
        self.graph.vs["count"] = list(map(
            operator.add, self.graph.vs["raw_count"],
            self.graph.vs["prop_count"]
        ))

    def best_leaves(self, thresh, retrieve="name"):
        subgraph = self.graph.subgraph(self.graph.vs.select(count_ge=thresh))
        leaves, max_count = [], 0
        for leaf in subgraph.vs.select(lambda v: v.indegree() == 0):
            if leaf["count"] > max_count:
                max_count = leaf["count"]
                leaves = [leaf[retrieve]]
            elif leaf["count"] == max_count:
                leaves.append(leaf[retrieve])
        return leaves


class DataDict(collections.OrderedDict):

    def shuffle(self, random_state=np.random):
        shuffled = DataDict()
        shuffle_idx = None
        for item in self:
            shuffle_idx = random_state.permutation(self[item].shape[0]) \
                if shuffle_idx is None else shuffle_idx
            shuffled[item] = self[item][shuffle_idx]
        return shuffled

    @property
    def size(self):
        data_size = set([item.shape[0] for item in self.values()])
        assert len(data_size) == 1
        return data_size.pop()

    @property
    def shape(self):  # Compatibility with numpy arrays
        return [self.size]

    def __getitem__(self, fetch):
        if isinstance(fetch, (slice, np.ndarray)):
            return DataDict([
                (item, self[item][fetch]) for item in self
            ])
        return super(DataDict, self).__getitem__(fetch)


def densify(arr):
    if scipy.sparse.issparse(arr):
        return arr.toarray()
    return arr


def empty_safe(fn, dtype):
    def _fn(x):
        if x.size:
            return fn(x)
        return x.astype(dtype)
    return _fn


decode = empty_safe(np.vectorize(lambda _x: _x.decode("utf-8")), str)
encode = empty_safe(np.vectorize(lambda _x: str(_x).encode("utf-8")), "S")
upper = empty_safe(np.vectorize(lambda x: str(x).upper()), str)
lower = empty_safe(np.vectorize(lambda x: str(x).lower()), str)
tostr = empty_safe(np.vectorize(str), str)


# training
from .helpers import *
from .evaluate import *
from torch.autograd import Variable

n_epochs = 200

def train(net, loader, n_epochs):

    n_batches = len(loader)
    for e in range(1, n_epochs + 1):
        iter_losses = []
        for i, data in enumerate(loader):
            batch = data
            try:
                batch_losses = net.train_step(batch, epoch=(e-1), it=i, n_batches=n_batches)
                #print(batch_losses)
            except Exception as e:
                print(f"Training stopped due to exception: {e}")
                return

            iter_losses.append(npy(batch_losses))

def train_cfg(net, loader, n_epochs, eval_data, batch_size, eval_interval):

    n_batches = len(loader)
    for e in range(1, n_epochs + 1):
        iter_losses = []
        for i, data in enumerate(loader):
            batch, _ = data
            try:
                batch_losses = net.train_step(batch, epoch=(e-1), it=i, n_batches=n_batches)
            except Exception as e:
                print(f"Training stopped due to exception: {e}")
                return

            iter_losses.append(npy(batch_losses))
        logs = get_logs(net, eval_data=eval_data, batch_size=batch_size, eval_interval=eval_interval,
                        iter_losses=iter_losses, epoch=e, include_params=True)
        print(logs)


import scanpy as sc
import anndata as ad
import numpy as np
import pandas as pd

def geneactivity(adata,
                 gtf_file,
                 key_added='gene',
                 upstream=10000,
                 feature_type='gene',
                 annotation='ENSEMBL',
                 layer_name='geneactivity',
                 raw=False,
                 maohao=False):
    """
    
    Build an AnnData object containing the number of open features
    (windows, peaks, etc) overlapping genes (gene bodies +
    5kb upstream of TSS).  
    It is possible to extend the distance from the TSS with the upstream parameter.
    GTF files can contain multiple annotations (HAVANA, ENSEMBL, etc.). 
    Rather than using multiple time the same gene
    It is possible to specify which genome annotation is desired using the parameter annotation.
    The annotation file can contain more than... 
    
    INPUT
    -----
    adata : input AnnData
    gtf_file : input gtf file name + path
    key_added :
    upstream :
    featyre_type
    transcripts or genes
    
    """
    ### extracting the genes
    gtf = {}
    with open(gtf_file) as f:
        for line in f:
            if line[0:2] != '##' and '\t'+feature_type+'\t' in line and '\t'+annotation+'\t' in line:
                line = line.rstrip('\n').split('\t')
                if line[6] == '-':
                    if line[0] not in gtf.keys():
                        gtf[line[0]] = [[int(line[3]), int(line[4])+upstream,line[-1].split(';')[:-1]]]
                    else:
                        gtf[line[0]].append([int(line[3]), int(line[4])+upstream,line[-1].split(';')[:-1]])
                else:
                    if line[0] not in gtf.keys():
                        gtf[line[0]] = [[int(line[3])-upstream, int(line[4]),line[-1].split(';')[:-1]]]
                    else:
                        gtf[line[0]].append([int(line[3])-upstream, int(line[4]),line[-1].split(';')[:-1]])

    # extracting the feature coordinates
    if raw==True:
        raw_adata = adata.raw.to_adata()
    else:
        raw_adata = adata.copy()
    raw_adata_features = {}
    feature_index = 0
    for line in raw_adata.var_names:
        if maohao:
            tmp = []
            tmp_ = line.split(':')
            tmp.append(tmp_[0])
            tmp_ = tmp_[1].split('-')
            tmp.append(tmp_[0])
            tmp.append(tmp_[1])
            line = tmp
        else:
            line = line.split('-')
        if line[0] not in raw_adata_features.keys():
            raw_adata_features[line[0]] = [[int(line[1]),int(line[2]), feature_index]]
        else:
            raw_adata_features[line[0]].append([int(line[1]),int(line[2]), feature_index])
        feature_index += 1
    
    ## find the features overlaping the genes. and build the count matrix
    gene_index = []
    gene_activity_X = []
 
    for chrom in gtf.keys():
        if chrom in raw_adata_features.keys():
            #print(chrom)
            chrom_index = 0
            previous_features_index = 0
            for gene in gtf[chrom]:
                gene_values = []
                gene_start = gene[0]
                gene_end = gene[1]
                for feature in raw_adata_features[chrom]:
                    feature_index = 0
                    if (feature[1]<= gene_start): # the window is before the gene. we need to test the next window.
                        continue
                    elif (gene_end <= feature[0]): # the window is after the gene. we need totest the next gene.
                        break
                    else: # the window is overlapping the gene. 
                        gene_values.append(raw_adata.X[:,feature[2]])
                if gene_values != []:
                    gene_activity_X.append(np.sum(gene_values, axis=0))
                    gene_index.append(gene[-1])
           
    gene_activity_X = np.transpose(np.array(gene_activity_X))#np.concatenate(tuple(gene_activity_X), axis=-1)

    # get the variable metadata
    gene_name = []
    if feature_type=='transcript':
        for x in gene_index:
            for y in x:
                if 'transcript_name' in y:
                    gene_name.append(y.lstrip(' transcript_name "').rstrip('"'))
        #gene_name = [x[7].lstrip(' transcript_name "').rstrip('"') for x in gene_index]
    elif feature_type=='gene':
        for x in gene_index:
            for y in x:
                if 'gene_name' in y:
                    gene_name.append(y.lstrip(' gene_name "').rstrip('"'))
        #gene_name = [x[4].lstrip(' gene_name "').rstrip('"') for x in gene_index]

    metadata_genes = {'gene_id' : [],
                      'transcript_id' : [],
                      'gene_type' : [],
                      'gene_name' : [],
                      'transcript_type' : [],
                      'transcript_name' : [],
                      'protein_id' : []}

    for line in gene_index:
        dico_line = {}
        for element in line:
            if ' "' in element:
                dico_line[element.rstrip('"').lstrip(" ").split(' "')[0]] = element.rstrip('"').lstrip(" ").split(' "')[1]
    
        for key in metadata_genes.keys():
            if key in dico_line.keys():
                metadata_genes[key].append(dico_line[key])
            else:
                metadata_genes[key].append('NA')  
            
    dataframe_genes = pd.DataFrame.from_dict(metadata_genes)
    dataframe_genes.index = gene_name

    #adata.layers[layer_name] = ad.AnnData(gene_activity_X, var=dataframe_genes, obs=raw_adata.obs)
    gene_adata = ad.AnnData(gene_activity_X, var=dataframe_genes, obs=raw_adata.obs)
    gene_adata.uns = adata.uns.copy()
    gene_adata.obsm = adata.obsm.copy()
    gene_adata.obsp = adata.obsp.copy()
    return(gene_adata)


# assembling the whole simple/contrastive MVC model
from moclust.multiview_autoencoders import *
from .loss import *
from .fusion import *
from .ddc import *
from .optimizer import *

import torch.nn as nn

def he_init_weights(module):
    """
    Initialize network weights using the He (Kaiming) initialization strategy.

    :param module: Network module
    :type module: nn.Module
    """
    if isinstance(module, (nn.Conv2d, nn.Linear)):
        nn.init.kaiming_normal_(module.weight)

class ModelBase(nn.Module):
    def __init__(self):
        """
        Model base class
        """
        super().__init__()

        self.fusion = None
        self.optimizer = None
        self.loss = None

    def calc_losses(self, ignore_in_total=tuple()):
        return self.loss(self, ignore_in_total=ignore_in_total)

    def train_step(self, batch, epoch, it, n_batches):
        self.optimizer.zero_grad()
        _ = self(batch)
        losses = self.calc_losses()
        losses["tot"].backward()
        self.optimizer.step(epoch + it / n_batches)
        return losses

# simple MVC for single cell multiomics data
class scMVC_simple(ModelBase):

    def __init__(self, cfg):

        super().__init__()

        self.cfg = cfg
        self.output = self.hidden = self.fused = None

        # Define Backbones and Fusion modules
        self.mvae = multiview_autoencoders(cfg.multiview_encoders_config)
        # print(self.multiview_encoders.output_sizes)
        self.fusion = get_fusion_module(cfg.fusion_config, self.mvae.output_sizes)
        # Define clustering module
        self.ddc = DDC(input_dim=self.fusion.output_size, cfg=cfg.cm_config)
        # Define loss-module
        self.loss = Loss(cfg=cfg.loss_config)
        # Initialize weights.
        self.apply(he_init_weights)

        # Instantiate optimizer
        self.optimizer = Optimizer(cfg.optimizer_config, self.parameters())

    def forward(self, mv_input):
        self.mv_input = mv_input
        self.output_mu = self.mvae.forward(mv_input)[0]
        self.output_mean = self.mvae.forward(mv_input)[1]
        self.output_disp = self.mvae.forward(mv_input)[2]
        self.output_pi = self.mvae.forward(mv_input)[3]
        #print(self.multiview_encoders_outputs)
        self.fused = self.fusion(self.output_mu)
        self.output_soft_label, self.hidden = self.ddc(self.fused)
        self.output = [self.output_soft_label, self.output_mu, self.output_mean, self.output_disp, self.output_pi]
        return self.output

#contrastive MVC for single cell multiomics data
class scMVC_contrast(ModelBase):
    def __init__(self, cfg):
        """
        Implementation of the CoMVC model.

        :param cfg: Model config. See `config.defaults.CoMVC` for documentation on the config object.
        """
        super().__init__()

        self.cfg = cfg
        self.output = self.hidden = self.fused = self.projections = None

        # Define Backbones and Fusion modules
        self.mvae = multiview_autoencoders(cfg.multiview_encoders_config)
        # print(self.multiview_encoders.output_sizes)
        self.fusion = get_fusion_module(cfg.fusion_config, self.mvae.output_sizes)

        mvae_sizes = self.mvae.output_sizes
        assert all([mvae_sizes[0] == s for s in mvae_sizes]), f"CoMVC requires all encoders to have the same " \
                                                          f"output size. Got: {mvae_sizes}"

        if cfg.projector_config is None:
            self.projector = nn.Identity()
        else:
            self.projector = MLP(cfg.projector_config, input_size=mvae_sizes[0])

        # Define clustering module
        self.ddc = DDC(input_dim=self.fusion.output_size, cfg=cfg.cm_config)
        # Define loss-module
        self.loss = Loss(cfg=cfg.loss_config)
        # Initialize weights.
        self.apply(he_init_weights)
        # Instantiate optimizer
        self.optimizer = Optimizer(cfg.optimizer_config, self.parameters())

    def forward(self, mv_input):
        self.mv_input = mv_input
        self.output_mu = self.mvae.forward(mv_input)[0]
        self.output_mean = self.mvae.forward(mv_input)[1]
        self.output_disp = self.mvae.forward(mv_input)[2]
        self.output_pi = self.mvae.forward(mv_input)[3]
        #print(self.multiview_encoders_outputs)
        self.fused = self.fusion(self.output_mu)
        self.projections = self.projector(th.cat(self.output_mu, dim=0))
        self.output_soft_label, self.hidden = self.ddc(self.fused)
        self.output = [self.output_soft_label, self.output_mu, self.output_mean, self.output_disp, self.output_pi, self.fused, self.hidden]
        return self.output


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="MoClust",
    version="0.0.4",
    author="yuanmusu",
    author_email="yuanmusu@pku.edu.cn",
    description="A small example package",
    long_description="A deep learning-based clustering method for single-cell multi-omics data",
    long_description_content_type="text/markdown",
    url="https://github.com/ddb-qiwang/MoClust",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    python_requires='>=3.8',
    install_requires =[
        "scanpy >= 1.6.0",
        "numpy >= 1.21.6",
        "pandas >= 1.3.5",
        "torch == 1.10.2",
        "scikit-learn >= 1.0.2",
        "scipy >= 1.4.1",
        "seaborn >= 0.9.0",
        "tabulate == 0.8.9",
        "pydantic == 1.10.2",
        "typing >= 3.5.0",]
)



import os
import argparse
import numpy as np
import torch as th
import matplotlib.pyplot as plt
import seaborn as sns
from tabulate import tabulate
from sklearn.metrics import normalized_mutual_info_score, adjusted_rand_score

from .helpers import *

IGNORE_IN_TOTAL = ("contrast",)


def calc_metrics(labels, pred):
    """
    Compute metrics.

    :param labels: Label tensor
    :type labels: th.Tensor
    :param pred: Predictions tensor
    :type pred: th.Tensor
    :return: Dictionary containing calculated metrics
    :rtype: dict
    """
    acc, cmat = ordered_cmat(labels, pred)
    metrics = {
        "acc": acc,
        "cmat": cmat,
        "nmi": normalized_mutual_info_score(labels, pred, average_method="geometric"),
        "ari": adjusted_rand_score(labels, pred),
    }
    return metrics


def get_log_params(net):
    """
    Get the network parameters we want to log.

    :param net: Model
    :type net:
    :return:
    :rtype:
    """
    params_dict = {}
    weights = []
    if getattr(net, "fusion", None) is not None:
        with th.no_grad():
            weights = net.fusion.get_weights(softmax=True)

    elif hasattr(net, "attention"):
        weights = net.weights

    for i, w in enumerate(npy(weights)):
        params_dict[f"fusion/weight_{i}"] = w

    if hasattr(net, "discriminators"):
        for i, discriminator in enumerate(net.discriminators):
            d0, dv = npy([discriminator.d0, discriminator.dv])
            params_dict[f"discriminator_{i}/d0/mean"] = d0.mean()
            params_dict[f"discriminator_{i}/d0/std"] = d0.std()
            params_dict[f"discriminator_{i}/dv/mean"] = dv.mean()
            params_dict[f"discriminator_{i}/dv/std"] = dv.std()

    return params_dict


def get_eval_data(dataset, n_eval_samples, batch_size):
    """
    Create a dataloader to use for evaluation

    :param dataset: Inout dataset.
    :type dataset: th.utils.data.Dataset
    :param n_eval_samples: Number of samples to include in the evaluation dataset. Set to None to use all available
                           samples.
    :type n_eval_samples: int
    :param batch_size: Batch size used for training.
    :type batch_size: int
    :return: Evaluation dataset loader
    :rtype: th.utils.data.DataLoader
    """
    if n_eval_samples is not None:
        *views, labels = dataset.tensors
        n = views[0].size(0)
        idx = np.random.choice(n, min(n, n_eval_samples), replace=False)
        views, labels = [v[idx] for v in views], labels[idx]
        dataset = th.utils.data.TensorDataset(*views, labels)

    eval_loader = th.utils.data.DataLoader(dataset, batch_size=int(batch_size), shuffle=True, num_workers=0,
                                           drop_last=False, pin_memory=False)
    return eval_loader


def batch_predict(net, eval_data, batch_size, if_train=True, if_latent=False, if_recon=False, if_softlabel=False):
    """
    Compute predictions for `eval_data` in batches. Batching does not influence predictions, but it influences the loss
    computations.

    :param net: Model
    :type net:
    :param eval_data: Evaluation dataloader
    :type eval_data: th.utils.data.DataLoader
    :param batch_size: Batch size
    :type batch_size: int
    :return: Label tensor, predictions tensor, list of dicts with loss values, array containing mean and std of cluster
             sizes.
    :rtype:
    """
    input_x = []
    input_y = []
    predictions = []
    softlabel = []
    labels = []
    losses = []
    cluster_sizes = []
    latent_features = []
    fused_features = []
    hidden_features = []
    totalmeanx = []
    totalmeany = []
    totaldispx = []
    totaldispy = []
    totalpix = []
    totalpiy = []

    net.eval()
    with th.no_grad():
        for i, (batch, label) in enumerate(eval_data):
            pred = net(batch)[0]
            latent = net(batch)[1]
            mean = net(batch)[2]
            disp = net(batch)[3]
            pi = net(batch)[4]
            fused = net(batch)[5]
            hidden = net(batch)[6]
            input_x.append(npy(batch[0][0]))
            input_y.append(npy(batch[0][1]))
            labels.append(npy(label))
            softlabel.append(npy(pred))
            predictions.append(npy(pred).argmax(axis=1))
            latent_features.append(npy(latent))
            fused_features.append(npy(fused))
            hidden_features.append(npy(hidden))
            totalmeanx.append(npy(mean[0]))
            totalmeany.append(npy(mean[1]))
            totaldispx.append(npy(disp[0]))
            totaldispy.append(npy(disp[1]))
            totalpix.append(npy(pi[0]))
            totalpiy.append(npy(pi[1]))

            # Only calculate losses for full batches
            if label.size(0) == batch_size:
                batch_losses = net.calc_losses(ignore_in_total=IGNORE_IN_TOTAL)
                losses.append(npy(batch_losses))
                cluster_sizes.append(npy(pred.sum(dim=0)))

    input_x = np.concatenate(input_x, axis=0)
    input_y = np.concatenate(input_y, axis=0)
    labels = np.concatenate(labels, axis=0)
    predictions = np.concatenate(predictions, axis=0)
    softlabel = np.concatenate(softlabel, axis=0)
    latent_features = np.concatenate(latent_features, axis=1)
    fused_features = np.concatenate(fused_features, axis=0)
    hidden_features = np.concatenate(hidden_features, axis=0)
    totalmeanx = np.concatenate(totalmeanx, axis=0)
    totaldispx = np.concatenate(totaldispx, axis=0)
    totalpix = np.concatenate(totalpix, axis=0)
    totalmeany = np.concatenate(totalmeany, axis=0)
    totaldispy = np.concatenate(totaldispy, axis=0)
    totalpiy = np.concatenate(totalpiy, axis=0)
    if if_recon:
        return input_x, input_y, totalmeanx, totalmeany, totaldispx, totaldispy, totalpix, totalpiy
    if if_latent:
        return labels, predictions, latent_features, fused_features, hidden_features
    if if_softlabel:
        return softlabel
    if if_train:
        net.train()
    return labels, predictions, losses, np.array(cluster_sizes).sum(axis=0)

def batch_predict_nolabel(net, eval_data, if_train=True, if_latent=False, if_recon=False, if_softlabel=False):
    """
    Compute predictions for `eval_data` in batches. Batching does not influence predictions, but it influences the loss
    computations.

    :param net: Model
    :type net:
    :param eval_data: Evaluation dataloader
    :type eval_data: th.utils.data.DataLoader
    :param batch_size: Batch size
    :type batch_size: int
    :return: Label tensor, predictions tensor, list of dicts with loss values, array containing mean and std of cluster
             sizes.
    :rtype:
    """
    input_x = []
    input_y = []
    predictions = []
    softlabel = []
    losses = []
    cluster_sizes = []
    latent_features = []
    fused_features = []
    hidden_features = []
    totalmeanx = []
    totalmeany = []
    totaldispx = []
    totaldispy = []
    totalpix = []
    totalpiy = []

    net.eval()
    with th.no_grad():
        for i, batch in enumerate(eval_data):
            pred = net(batch)[0]
            latent = net(batch)[1]
            mean = net(batch)[2]
            disp = net(batch)[3]
            pi = net(batch)[4]
            fused = net(batch)[5]
            hidden = net(batch)[6]
            input_x.append(npy(batch[0][0]))
            input_y.append(npy(batch[0][1]))
            softlabel.append(npy(pred))
            predictions.append(npy(pred).argmax(axis=1))
            latent_features.append(npy(latent))
            fused_features.append(npy(fused))
            hidden_features.append(npy(hidden))
            totalmeanx.append(npy(mean[0]))
            totalmeany.append(npy(mean[1]))
            totaldispx.append(npy(disp[0]))
            totaldispy.append(npy(disp[1]))
            totalpix.append(npy(pi[0]))
            totalpiy.append(npy(pi[1]))

    if if_recon:
        return input_x, input_y, totalmeanx, totalmeany, totaldispx, totaldispy, totalpix, totalpiy
    if if_latent:
        return predictions, latent_features, fused_features, hidden_features
    if if_softlabel:
        return softlabel
    if if_train:
        net.train()
    return predictions, losses, np.array(cluster_sizes).sum(axis=0)

def get_logs(net, eval_data, batch_size, eval_interval, iter_losses=None, epoch=None, include_params=True):
    if iter_losses is not None:
        logs = add_prefix(dict_means(iter_losses), "iter_losses")
    else:
        logs = {}
    if (epoch is None) or ((epoch % eval_interval) == 0):
        labels, pred, eval_losses, cluster_sizes = batch_predict(net, eval_data, batch_size)
        eval_losses = dict_means(eval_losses)
        logs.update(add_prefix(eval_losses, "eval_losses"))
        logs.update(add_prefix(calc_metrics(labels, pred), "metrics"))
        logs.update(add_prefix({"mean": cluster_sizes.mean(), "sd": cluster_sizes.std()}, "cluster_size"))
    if include_params:
        logs.update(add_prefix(get_log_params(net), "params"))
    if epoch is not None:
        logs["epoch"] = epoch
    return logs


def eval_run(cfg, cfg_name, experiment_identifier, run, net, eval_data, callbacks=tuple(), load_best=True):
    """
    Evaluate a training run.

    :param cfg: Experiment config
    :type cfg: config.defaults.Experiment
    :param cfg_name: Config name
    :type cfg_name: str
    :param experiment_identifier: 8-character unique identifier for the current experiment
    :type experiment_identifier: str
    :param run: Run to evaluate
    :type run: int
    :param net: Model
    :type net:
    :param eval_data: Evaluation dataloder
    :type eval_data: th.utils.data.DataLoader
    :param callbacks: List of callbacks to call after evaluation
    :type callbacks: List
    :param load_best: Load the "best.pt" model before evaluation?
    :type load_best: bool
    :return: Evaluation logs
    :rtype: dict
    """
    if load_best:
        model_path = get_save_dir(cfg_name, experiment_identifier, run) / "best.pt"
        if os.path.isfile(model_path):
            net.load_state_dict(th.load(model_path))
        else:
            print(f"Unable to load best model for evaluation. Model file not found: {model_path}")
    logs = get_logs(cfg, net, eval_data, include_params=True)
    for cb in callbacks:
        cb.at_eval(net=net, logs=logs)
    return logs


def plot_projection(X, method, hue, ax, title=None, cmap="tab10", legend_title=None, legend_loc=1, **kwargs):
    X = project(X, method)
    pl = sns.scatterplot(x=X[:, 0], y=X[:, 1], hue=hue, ax=ax, legend="full", palette=cmap, **kwargs)
    leg = pl.get_legend()
    leg._loc = legend_loc
    if title is not None:
        ax.set_title(title)
    if legend_title is not None:
        leg.set_title(legend_title)


def project(X, method):
    if method == "pca":
        from sklearn.decomposition import PCA
        return PCA(n_components=2).fit_transform(X)
    elif method == "tsne":
        from sklearn.manifold import TSNE
        return TSNE(n_components=2).fit_transform(X)
    elif method is None:
        return X
    else:
        raise RuntimeError()


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", dest="cfg_name", required=True)
    parser.add_argument("-t", "--tag", dest="tag", required=True)
    parser.add_argument("--plot", action="store_true")
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()
    eval_experiment(args.cfg_name, args.tag, args.plot)



import math
import pandas as pd
import numpy as np
import torch.nn as nn
from sklearn.metrics import confusion_matrix
from scipy.optimize import linear_sum_assignment


def npy(t, to_cpu=True):
    """
    Convert a tensor to a numpy array.

    :param t: Input tensor
    :type t: th.Tensor
    :param to_cpu: Call the .cpu() method on `t`?
    :type to_cpu: bool
    :return: Numpy array
    :rtype: np.ndarray
    """
    if isinstance(t, (list, tuple)):
        # We got a list. Convert each element to numpy
        return [npy(ti) for ti in t]
    elif isinstance(t, dict):
        # We got a dict. Convert each value to numpy
        return {k: npy(v) for k, v in t.items()}
    # Assuming t is a tensor.
    if to_cpu:
        return t.cpu().detach().numpy()
    return t.detach().numpy()


def ensure_iterable(elem, expected_length=1):
    if isinstance(elem, (list, tuple)):
        assert len(elem) == expected_length, f"Expected iterable {elem} with length {len(elem)} does not have " \
                                             f"expected length {expected_length}"
    else:
        elem = expected_length * [elem]
    return elem


def dict_means(dicts):
    """
    Compute the mean value of keys in a list of dicts

    :param dicts: Input dicts
    :type dicts: List[dict]
    :return: Mean values
    :rtype: dict
    """
    return pd.DataFrame(dicts).mean(axis=0).to_dict()


def add_prefix(dct, prefix, sep="/"):
    """
    Add a prefix to all keys in `dct`.

    :param dct: Input dict
    :type dct: dict
    :param prefix: Prefix
    :type prefix: str
    :param sep: Separator between prefix and key
    :type sep: str
    :return: Dict with prefix prepended to all keys
    :rtype: dict
    """
    return {prefix + sep + key: value for key, value in dct.items()}


def ordered_cmat(labels, pred):
    """
    Compute the confusion matrix and accuracy corresponding to the best cluster-to-class assignment.

    :param labels: Label array
    :type labels: np.array
    :param pred: Predictions array
    :type pred: np.array
    :return: Accuracy and confusion matrix
    :rtype: Tuple[float, np.array]
    """
    cmat = confusion_matrix(labels, pred)
    ri, ci = linear_sum_assignment(-cmat)
    ordered = cmat[np.ix_(ri, ci)]
    acc = np.sum(np.diag(ordered))/np.sum(ordered)
    return acc, ordered

def he_init_weights(module):
    """
    Initialize network weights using the He (Kaiming) initialization strategy.

    :param module: Network module
    :type module: nn.Module
    """
    if isinstance(module, (nn.Conv2d, nn.Linear)):
        nn.init.kaiming_normal_(module.weight)


def num2tuple(num):
    return num if isinstance(num, (tuple, list)) else (num, num)


def conv2d_output_shape(h_w, kernel_size=1, stride=1, pad=0, dilation=1):
    """
    Compute the output shape of a convolution operation.

    :param h_w: Height and width of input
    :type h_w: Tuple[int, int]
    :param kernel_size: Size of kernel
    :type kernel_size: Union[int, Tuple[int, int]]
    :param stride: Stride of convolution
    :type stride: Union[int, Tuple[int, int]]
    :param pad: Padding (in pixels)
    :type pad: Union[int, Tuple[int, int]]
    :param dilation: Dilation
    :type dilation: Union[int, Tuple[int, int]]
    :return: Height and width of output
    :rtype: Tuple[int, int]
    """
    h_w, kernel_size, stride, = num2tuple(h_w), num2tuple(kernel_size), num2tuple(stride)
    pad, dilation = num2tuple(pad), num2tuple(dilation)

    h = math.floor((h_w[0] + 2 * pad[0] - dilation[0] * (kernel_size[0] - 1) - 1) / stride[0] + 1)
    w = math.floor((h_w[1] + 2 * pad[1] - dilation[1] * (kernel_size[1] - 1) - 1) / stride[1] + 1)
    return h, w
